import uuid

import requests

from django.db import Error
from django.utils import timezone
from eveuniverse.models import EveMarketPrice, EveType, EveTypeMaterial

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.app_settings import BUYBACKPROGRAM_TRACKING_PREFILL
from buybackprogram.constants import (
    BLUE_LOOT_TYPE_IDS,
    MOON_ORE_EVE_GROUPS,
    OPE_EVE_GROUPS,
    ORE_EVE_GROUPS,
    RED_LOOT_TYPE_IDS,
)
from buybackprogram.models import ItemPrices, ProgramItem, Tracking, TrackingItem
from buybackprogram.notes import (
    note_compressed_price_used,
    note_item_disallowed,
    note_item_specific_tax,
    note_missing_jita_buy,
    note_missing_typematerials,
    note_no_price_data,
    note_npc_price,
    note_price_dencity_tax,
    note_refined_price_used,
    note_unpublished_item,
)

logger = get_extension_logger(__name__)


def get_item_tax(program, item_id):
    try:
        program_item_settings = ProgramItem.objects.get(
            program=program, item_type__id=item_id
        )

        return program_item_settings.item_tax

    except ProgramItem.DoesNotExist:
        return False


def use_npc_price(item, program):
    if item.id in BLUE_LOOT_TYPE_IDS and program.blue_loot_npc_price:
        return True
    elif item.id in RED_LOOT_TYPE_IDS and program.red_loot_npc_price:
        return True
    elif item.eve_group.id in OPE_EVE_GROUPS and program.ope_npc_price:
        return True
    else:
        return False


def get_or_create_prices(item_id):

    try:
        return ItemPrices.objects.get(eve_type_id=item_id)

    except ItemPrices.DoesNotExist:

        response_fuzzwork = requests.get(
            "https://market.fuzzwork.co.uk/aggregates/",
            params={
                "types": item_id,
                "station": 60003760,
            },
        )

        items_fuzzwork = response_fuzzwork.json()

        buy = int(float(items_fuzzwork[str(item_id)]["buy"]["max"]))
        sell = int(float(items_fuzzwork[str(item_id)]["sell"]["min"]))
        updated = timezone.now()

        try:
            price = ItemPrices.objects.create(
                eve_type_id=item_id, buy=buy, sell=sell, updated=updated
            )

            return price
        except Error as e:
            logger.error("Error updating prices: %s" % e)


def get_npc_price(item_id):
    try:
        return EveMarketPrice.objects.get(eve_type=item_id)
    except Error as e:
        logger.error("Error getting NPC prices for %s: %s" % (item_id, e))


def getList(dict):
    return dict.keys()


def is_ore(item_id):
    if item_id in ORE_EVE_GROUPS:
        return True
    else:
        return False


def is_moon_ore(item_id):
    if item_id in MOON_ORE_EVE_GROUPS:
        return True
    else:
        return False


def has_buy_price(item):
    if (
        not item["raw_prices"]
        and not item["material_prices"]
        and not item["compression_prices"]
        and not item["npc_prices"]
    ):
        return False
    else:
        return True


def get_price_dencity_tax(program, item_value, item_volume, item_quantity):
    # If price dencity tax should be applied
    if program.price_dencity_modifier:

        if not item_volume <= 0:
            item_isk_dencity = item_value / item_volume
        else:
            item_isk_dencity = False

        logger.debug(
            "Values: Our item isk dencity is at %s ISK/m³ with value: %s, volume: %s"
            % (item_isk_dencity, item_value, item_volume)
        )

        if item_isk_dencity < program.price_dencity_treshold and item_isk_dencity:
            logger.debug(
                "Isk dencity under threshold value, applying extra taxes of %s"
                % program.price_dencity_tax
            )

            return program.price_dencity_tax
        else:
            logger.debug("Isk dencity above threshold value, no extra taxes added")
            return False
    else:
        return False


# This method will get the price information for the item. It will not calculate the values as in price including taxes.
def get_item_prices(item_type, name, quantity, program):

    notes = []
    has_price_variants = False

    # Get special taxes and see if our item belongs to this table

    program_item_settings = ProgramItem.objects.filter(
        program=program, item_type__id=item_type.id
    ).first()

    # Get item raw price information
    item_price = get_or_create_prices(item_type.id)

    # Check what items are allowed
    if program.allow_all_items:
        if program_item_settings:
            item_disallowed = program_item_settings.disallow_item

            notes.append(note_item_disallowed(item_disallowed, name))

        else:
            item_disallowed = False
    else:
        if program_item_settings:
            item_disallowed = program_item_settings.disallow_item

            notes.append(note_item_disallowed(item_disallowed, name))

        else:
            item_disallowed = True

            notes.append(note_item_disallowed(item_disallowed, name))

    # If item is somehow not published (expired items etc.)
    if not item_type.published:
        item_disallowed = True

        notes.append(note_unpublished_item(name))

    if not item_disallowed:

        # If raw ore value should not be taken into account
        if not program.use_raw_ore_value and is_ore(item_type.eve_group.id):
            item_raw_price = {
                "id": item_type.id,
                "quantity": quantity,
                "buy": False,
                "sell": False,
            }

        else:
            item_raw_price = {
                "id": item_type.id,
                "quantity": quantity,
                "buy": item_price.buy,
                "sell": item_price.sell,
            }

        notes.append(note_missing_jita_buy(item_price.buy, name))

        # Check if we should get refined value for the item
        if (
            is_ore(item_type.eve_group.id) or is_moon_ore(item_type.eve_group.id)
        ) and program.use_refined_value:
            item_material_price = []
            # Get all refining materials for item

            type_materials = EveTypeMaterial.objects.filter(
                eve_type_id=item_type.id
            ).prefetch_related("eve_type")

            notes.append(note_missing_typematerials(type_materials, name))

            # Get price details for the materials inside the item
            for material in type_materials:
                material_price = get_or_create_prices(material.material_eve_type.id)

                # Quantity of refined materials
                material_quantity = (
                    material.quantity * quantity
                ) / item_type.portion_size

                material_type_prices = {
                    "id": material.material_eve_type.id,
                    "quantity": material_quantity,
                    "unit_quantity": material.quantity,
                    "buy": material_price.buy,
                    "sell": material_price.sell,
                }

                item_material_price.append(material_type_prices)

            has_price_variants = True

            logger.debug("Prices: Got refined values for %s" % name)

        else:
            item_material_price = False
            type_materials = False
            logger.debug("Prices: No refined value used for %s" % name)

        # Get compressed versions of the ores that are not yet compressed
        if is_ore(item_type.eve_group.id) and program.use_compressed_value:

            if "Compressed" in name:
                compresed_name = name
            else:
                compresed_name = "Compressed " + name

            compresed_type = EveType.objects.filter(name=compresed_name).first()
            compression_ratio = item_type.portion_size

            compression_price = get_or_create_prices(compresed_type.id)

            logger.debug(
                "Got prices %s ISK for %s" % (compression_price.buy, compresed_name)
            )

            compressed_type_prices = {
                "id": compression_price.eve_type_id,
                "quantity": quantity / compression_ratio,
                "buy": compression_price.buy,
                "sell": compression_price.sell,
            }

            has_price_variants = True

            logger.debug(
                "Prices: Got compression prices for %s based on original item %s"
                % (compresed_name, name)
            )

        # If item can't or should not be compressed
        else:
            logger.debug("Prices: No compression required/available for %s" % name)
            compressed_type_prices = False

        # Get NPC prices
        if use_npc_price(item_type, program):
            item_npc_price = get_npc_price(item_type.id)

            npc_type_prices = {
                "id": item_type.id,
                "quantity": quantity,
                "buy": item_npc_price.average_price,
                "sell": False,
            }

            logger.debug("Prices: Got NPC buy price for %s" % name)

        else:

            logger.debug("Prices: No NPC prices required/available for %s" % name)
            npc_type_prices = False

        prices = {
            "quantity": quantity,
            "type_prices": item_price,
            "raw_prices": item_raw_price,
            "material_prices": item_material_price,
            "compression_prices": compressed_type_prices,
            "npc_prices": npc_type_prices,
            "has_price_variants": has_price_variants,
            "notes": notes,
        }
    else:
        prices = {
            "quantity": quantity,
            "type_prices": item_price,
            "raw_prices": False,
            "material_prices": False,
            "compression_prices": False,
            "npc_prices": False,
            "has_price_variants": has_price_variants,
            "notes": notes,
        }

    return prices


def get_item_values(item_type, item_prices, program):

    type_value = False
    compression_value = False
    item_tax = False
    refined = []
    compressed = False
    type_raw_value = False
    compression_raw_value = False

    # Get values for the type prices (base prices)
    if item_prices["raw_prices"]:

        quantity = item_prices["raw_prices"]["quantity"]
        sell = item_prices["raw_prices"]["sell"]
        buy = item_prices["raw_prices"]["buy"]
        price = buy

        if not item_type.volume <= 0:
            price_dencity = price / item_type.volume
        else:
            price_dencity = False
        price_dencity_tax = get_price_dencity_tax(
            program, price, item_type.volume, quantity
        )
        program_tax = program.tax
        item_tax = get_item_tax(program, item_type.id)
        tax_multiplier = (100 - (program_tax + item_tax + price_dencity_tax)) / 100

        logger.debug("Values: Calculating type value for %s" % item_type)

        type_raw_value = quantity * price
        type_value = type_raw_value * tax_multiplier

        raw_item = {
            "id": item_type.id,
            "name": item_type.name,
            "quantity": quantity,
            "buy": buy,
            "sell": sell,
            "program_tax": program_tax,
            "item_tax": item_tax,
            "price_dencity_tax": price_dencity_tax,
            "total_tax": program_tax + item_tax + price_dencity_tax,
            "price_dencity": price_dencity,
            "unit_value": price * tax_multiplier,
            "raw_value": type_raw_value,
            "value": type_value,
            "is_buy_value": False,
            "notes": [],
        }

        raw_item["notes"].append(
            note_price_dencity_tax(raw_item["name"], price_dencity, price_dencity_tax)
        )
        raw_item["notes"].append(note_item_specific_tax(item_type.name, item_tax))

    else:
        raw_item = {
            "unite_value": False,
            "value": False,
            "total_tax": False,
            "raw_value": False,
        }

    # Get values for refined variant
    if item_prices["material_prices"]:

        refined = {
            "materials": [],
            "unit_value": False,
            "total_tax": False,
            "raw_value": False,
            "value": False,
            "is_buy_value": False,
        }

        material_count = len(item_prices["material_prices"])

        for material in item_prices["material_prices"]:

            materials = EveType.objects.filter(id=material["id"]).first()

            quantity = material["quantity"]
            sell = material["sell"]
            buy = material["buy"]
            price = buy
            price_dencity = price / materials.volume
            price_dencity_tax = get_price_dencity_tax(
                program, price, materials.volume, quantity
            )
            program_tax = program.tax
            item_tax = get_item_tax(program, item_type.id)
            refining_rate = program.refining_rate / 100
            tax_multiplier = (100 - (program_tax + item_tax + price_dencity_tax)) / 100

            raw_value = quantity * refining_rate * price
            value = raw_value * tax_multiplier

            r = {
                "id": material["id"],
                "name": materials.name,
                "quantity": quantity * refining_rate,
                "buy": buy,
                "sell": sell,
                "program_tax": program_tax,
                "item_tax": item_tax,
                "price_dencity_tax": price_dencity_tax,
                "total_tax": program_tax + item_tax + price_dencity_tax,
                "price_dencity": price_dencity,
                "unit_value": price * tax_multiplier,
                "raw_value": raw_value,
                "value": value,
                "notes": [],
            }

            r["notes"].append(
                note_price_dencity_tax(materials.name, price_dencity, price_dencity_tax)
            )
            r["notes"].append(note_item_specific_tax(materials.name, item_tax))

            refined["materials"].append(r)

            refined["value"] += value
            refined["raw_value"] += raw_value
            refined["total_tax"] += r["total_tax"] / material_count
            refined["unit_value"] += (
                price * tax_multiplier * material["unit_quantity"] * refining_rate
            )

    else:
        refined = {
            "value": False,
            "raw_value": False,
            "total_tax": False,
            "unit_value": False,
        }

    # Get values for compressed variant
    if item_prices["compression_prices"]:

        compressed_version = EveType.objects.filter(
            id=item_prices["compression_prices"]["id"]
        ).first()

        quantity = item_prices["compression_prices"]["quantity"]
        buy = item_prices["compression_prices"]["buy"]
        sell = item_prices["compression_prices"]["sell"]
        price = buy
        price_dencity = price / compressed_version.volume
        price_dencity_tax = get_price_dencity_tax(
            program, price, compressed_version.volume, quantity
        )
        program_tax = program.tax
        item_tax = get_item_tax(program, item_type.id)
        tax_multiplier = (100 - (program_tax + item_tax + price_dencity_tax)) / 100

        logger.debug("Values: Calculating compression value for %s" % item_type.id)

        compression_raw_value = quantity * price
        compression_value = compression_raw_value * tax_multiplier

        compressed = {
            "id": compressed_version.id,
            "name": compressed_version.name,
            "quantity": quantity,
            "buy": buy,
            "sell": sell,
            "program_tax": program_tax,
            "item_tax": item_tax,
            "price_dencity_tax": price_dencity_tax,
            "total_tax": program_tax + item_tax + price_dencity_tax,
            "price_dencity": price_dencity,
            "unit_value": price * tax_multiplier,
            "raw_value": compression_raw_value,
            "value": compression_value,
            "is_buy_value": False,
            "notes": [],
        }

        compressed["notes"].append(
            note_price_dencity_tax(
                compressed_version.name, price_dencity, price_dencity_tax
            )
        )
        compressed["notes"].append(
            note_item_specific_tax(compressed_version.name, item_tax)
        )
    else:
        compressed = {
            "value": False,
            "raw_value": False,
            "total_tax": False,
            "unit_value": False,
        }

    # Get value for NPC price
    if item_prices["npc_prices"]:

        quantity = item_prices["npc_prices"]["quantity"]
        sell = item_prices["npc_prices"]["sell"]
        buy = item_prices["npc_prices"]["buy"]
        price = buy

        if not item_type.volume <= 0:
            price_dencity = price / item_type.volume
        else:
            price_dencity = False
        price_dencity_tax = get_price_dencity_tax(
            program, price, item_type.volume, quantity
        )
        program_tax = program.tax
        item_tax = get_item_tax(program, item_type.id)
        tax_multiplier = (100 - (program_tax + item_tax + price_dencity_tax)) / 100

        logger.debug("Values: Calculating npc value for %s" % item_type)

        type_raw_value = quantity * price
        type_value = type_raw_value * tax_multiplier

        npc_item = {
            "id": item_type.id,
            "name": item_type.name,
            "quantity": quantity,
            "buy": buy,
            "sell": sell,
            "program_tax": program_tax,
            "item_tax": item_tax,
            "price_dencity_tax": price_dencity_tax,
            "total_tax": program_tax + item_tax + price_dencity_tax,
            "price_dencity": price_dencity,
            "unit_value": price * tax_multiplier,
            "raw_value": type_raw_value,
            "value": type_value,
            "is_buy_value": False,
            "notes": [],
        }

        npc_item["notes"].append(
            note_price_dencity_tax(npc_item["name"], price_dencity, price_dencity_tax)
        )
        npc_item["notes"].append(note_item_specific_tax(item_type.name, item_tax))

    else:
        npc_item = {
            "unite_value": False,
            "value": False,
            "total_tax": False,
            "raw_value": False,
        }

    # If there are no price variations at all for the item
    if not has_buy_price(item_prices):

        item_prices["notes"].append(note_no_price_data(item_type.name))

        raw_item = {
            "id": item_type.id,
            "name": item_type.name,
            "quantity": item_prices["quantity"],
            "buy": False,
            "sell": False,
            "program_tax": False,
            "item_tax": False,
            "price_dencity_tax": False,
            "total_tax": False,
            "price_dencity": False,
            "unit_value": False,
            "raw_value": False,
            "value": False,
            "is_buy_value": False,
        }

    # Get the highest value of the used pricing methods
    if not item_prices["npc_prices"]:
        buy_value = max([raw_item["value"], refined["value"], compressed["value"]])

        raw_value = max(
            [raw_item["raw_value"], refined["raw_value"], compressed["raw_value"]]
        )

        unit_value = max(
            [raw_item["unit_value"], refined["unit_value"], compressed["unit_value"]]
        )
    else:
        buy_value = npc_item["value"]
        raw_value = npc_item["raw_value"]
        unit_value = npc_item["unit_value"]

    # Determine what value we will use for buy value
    if buy_value == raw_item["value"]:

        raw_item["is_buy_value"] = True
        tax_value = raw_item["total_tax"]

        item_prices["notes"].append(
            note_price_dencity_tax(
                raw_item["name"],
                raw_item["price_dencity"],
                raw_item["price_dencity_tax"],
            )
        )
        item_prices["notes"].append(
            note_item_specific_tax(raw_item["name"], raw_item["item_tax"])
        )

    elif buy_value == refined["value"]:

        refined["is_buy_value"] = True
        tax_value = refined["total_tax"]

        item_prices["notes"].append(note_refined_price_used(raw_item["name"]))

        for material in refined["materials"]:
            item_prices["notes"].append(
                note_price_dencity_tax(
                    material["name"],
                    material["price_dencity"],
                    material["price_dencity_tax"],
                )
            )
            item_prices["notes"].append(
                note_item_specific_tax(
                    material["name"] + " material", material["item_tax"]
                )
            )

    elif buy_value == compressed["value"]:

        compressed["is_buy_value"] = True
        tax_value = compressed["total_tax"]

        item_prices["notes"].append(note_compressed_price_used(raw_item["name"]))

        item_prices["notes"].append(
            note_price_dencity_tax(
                compressed["name"],
                compressed["price_dencity"],
                compressed["price_dencity_tax"],
            )
        )
        item_prices["notes"].append(
            note_item_specific_tax(raw_item["name"], compressed["item_tax"])
        )
    elif buy_value == npc_item["value"]:

        npc_item["is_buy_value"] = True
        tax_value = npc_item["total_tax"]

        item_prices["notes"].append(note_npc_price(raw_item["name"]))

        item_prices["notes"].append(
            note_price_dencity_tax(
                raw_item["name"],
                raw_item["price_dencity"],
                raw_item["price_dencity_tax"],
            )
        )
        item_prices["notes"].append(
            note_item_specific_tax(raw_item["name"], raw_item["item_tax"])
        )

    logger.debug("Values: Best buy value for %s is %s ISK" % (item_type, buy_value))

    # Final values for this item
    values = {
        "name": item_type,
        "quantity": item_prices["quantity"],
        "normal": raw_item,
        "refined": refined,
        "compressed": compressed,
        "npc": npc_item,
        "type_value": raw_item["value"],
        "material_value": refined["value"],
        "compression_value": compressed["value"],
        "npc_value": npc_item["value"],
        "unit_value": unit_value,
        "raw_value": raw_value,
        "tax_value": tax_value,
        "buy_value": buy_value,
    }

    return values


def get_item_buy_value(buyback_data, program, donation):

    total_all_items = 0
    total_hauling_cost = 0
    contract_net_total = False
    total_donation = False
    tota_all_items_raw = 0

    # Get a grand total value of all buy prices
    for item in buyback_data:
        tota_all_items_raw += item["item_values"]["raw_value"]
        total_all_items += item["item_values"]["buy_value"]

    logger.debug(
        "Final: Total buy value for all items before expenses is %s ISK"
        % total_all_items
    )

    if donation > 0:

        total_donation = total_all_items * (donation / 100)

        logger.debug(
            "Seller will donate %s a total of %s ISK" % (donation, total_donation)
        )

    # Calculate hauling expenses
    if program.hauling_fuel_cost > 0:
        for item in buyback_data:
            if has_buy_price(item["item_prices"]):
                item_volume = item["type_data"].volume
                quantity = item["item_values"]["quantity"]
                hauling_cost = item_volume * program.hauling_fuel_cost * quantity

                total_hauling_cost += hauling_cost

                logger.debug(
                    "Final: Hauling cost %s m³ of %s is %s ISK"
                    % (
                        item["item_values"]["quantity"],
                        item["type_data"],
                        hauling_cost,
                    )
                )

        logger.debug(
            "Final: Total hauling cost for whole contract is %s ISK"
            % total_hauling_cost
        )

    contract_net_total = total_all_items - total_hauling_cost - total_donation

    logger.debug("Final: Net total after expenses is %s ISK" % contract_net_total)

    contract_net_prices = {
        "total_all_items_raw": tota_all_items_raw,
        "total_all_items": total_all_items,
        "total_tax_amount": tota_all_items_raw - total_all_items,
        "total_donation_amount": total_donation,
        "hauling_cost": program.hauling_fuel_cost,
        "total_hauling_cost": total_hauling_cost,
        "contract_net_total": contract_net_total,
    }

    return contract_net_prices


def item_missing(item_name, quantity):

    values = {
        "name": item_name,
        "quantity": quantity,
        "normal": False,
        "refined": False,
        "compressed": False,
        "type_value": False,
        "material_value": False,
        "compression_value": False,
        "unit_value": False,
        "raw_value": False,
        "buy_value": False,
    }

    return values


def get_tracking_number(
    user, program, form_donation, buyback_data, contract_price_data
):

    tracking_number = (
        BUYBACKPROGRAM_TRACKING_PREFILL + "-" + uuid.uuid4().hex[:6].upper()
    )

    logger.debug(
        "Contract net total for tracking %s is %s"
        % (tracking_number, contract_price_data["contract_net_total"])
    )

    tracking = Tracking(
        program=program,
        issuer_user=user,
        value=contract_price_data["total_all_items_raw"],
        taxes=contract_price_data["total_tax_amount"],
        hauling_cost=contract_price_data["total_hauling_cost"],
        donation=contract_price_data["total_donation_amount"],
        net_price=round(contract_price_data["contract_net_total"]),
        tracking_number=tracking_number,
    )

    tracking.save()

    objs = []

    for item in buyback_data:

        if item["type_data"]:
            tracking_item = TrackingItem(
                tracking=tracking,
                eve_type=item["type_data"],
                buy_value=item["item_values"]["unit_value"],
                quantity=item["item_values"]["quantity"],
            )

            objs.append(tracking_item)
    try:
        TrackingItem.objects.bulk_create(objs)
        logger.debug("Succesfully created items for tracking %s" % tracking_number)
    except Error as e:
        logger.error("Error creating tracking items: %s" % e)

    return tracking_number



from typing import Tuple

from django.contrib.auth.models import Group, User
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import Error, models
from django.utils.translation import gettext as _
from esi.errors import TokenExpiredError, TokenInvalidError
from esi.models import Token
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.authentication.models import CharacterOwnership, State

# Create your models here.
from allianceauth.eveonline.models import EveCorporationInfo
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.notification import (
    send_message_to_discord_channel,
    send_user_notification,
)
from buybackprogram.utils import get_site_url

from .decorators import fetch_token_for_owner
from .providers import esi

logger = get_extension_logger(__name__)


def get_sentinel_user():
    """
    get user or create one
    :return:
    """

    return User.objects.get_or_create(username="deleted")[0]


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_access", "Can access this app and see own statics."),
            (
                "manage_programs",
                "Can manage own buyback programs and see own program statics.",
            ),
            ("see_all_statics", "Can see all program statics."),
        )


class Owner(models.Model):
    """A corporation that has buyback programs"""

    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_ESI_UNAVAILABLE = 5

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
    ]

    corporation = models.ForeignKey(
        EveCorporationInfo,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    character = models.ForeignKey(
        CharacterOwnership,
        help_text="Character used for retrieving info",
        on_delete=models.deletion.PROTECT,
        related_name="+",
    )

    user = models.ForeignKey(
        User,
        help_text="User that manages the program",
        on_delete=models.deletion.PROTECT,
        related_name="+",
    )

    class Meta:
        default_permissions = ()

    @fetch_token_for_owner(
        [
            "esi-contracts.read_character_contracts.v1",
            "esi-contracts.read_corporation_contracts.v1",
        ]
    )
    def update_contracts_esi(self, token):

        logger.debug("Fetching contracts for %s" % self.user)

        # Get all contracts for owner
        contracts = self._fetch_contracts()

        logger.debug("Got %s character contracts" % len(contracts))

        # Get all contracts for owner corporation
        corporation_contracts = self._fetch_corporation_contracts()

        logger.debug("Got %s corporation contracts" % len(corporation_contracts))

        # Merge all contracts into a single list
        all_contracts = contracts + corporation_contracts

        logger.debug("Total contracts received: %s" % len(all_contracts))

        # Get all tracking objects from the database
        tracking_numbers = Tracking.objects.all()

        # Start looping for all stored tracking objects
        for tracking in tracking_numbers:
            # Start checking if we find any matches from our ESI contracts
            for contract in all_contracts:

                # Only get contracts with the correct prefill ticker
                if tracking.tracking_number in contract["title"]:

                    logger.debug(
                        "Found a matching contract from tracking for %s "
                        % contract["contract_id"]
                    )

                    # Check if we already have the contract stored
                    try:
                        old_contract = Contract.objects.get(
                            contract_id=contract["contract_id"]
                        )

                    except Contract.DoesNotExist:
                        logger.debug("No matching contracts found")
                        old_contract = Contract.objects.none()
                        old_contract.status = False

                    # If we have found a contract from database that is not yet finished
                    if old_contract.status not in ["finished", "rejected"]:

                        # Create or update the found contract
                        obj, created = Contract.objects.update_or_create(
                            contract_id=contract["contract_id"],
                            defaults={
                                "contract_id": contract["contract_id"],
                                "assignee_id": contract["assignee_id"],
                                "availability": contract["availability"],
                                "date_completed": contract["date_completed"],
                                "date_expired": contract["date_expired"],
                                "date_issued": contract["date_issued"],
                                "for_corporation": contract["for_corporation"],
                                "issuer_corporation_id": contract[
                                    "issuer_corporation_id"
                                ],
                                "issuer_id": contract["issuer_id"],
                                "start_location_id": contract["start_location_id"],
                                "price": contract["price"],
                                "status": contract["status"],
                                "title": contract["title"],
                                "volume": contract["volume"],
                            },
                        )

                        # If we have created a new contract
                        if created:
                            logger.debug(
                                "New contract %s created. Starting item fetch"
                                % contract["contract_id"]
                            )

                            character_id = self.character.character.character_id

                            # Get all items in the contract
                            contract_items = esi.client.Contracts.get_characters_character_id_contracts_contract_id_items(
                                character_id=character_id,
                                contract_id=contract["contract_id"],
                                token=token.valid_access_token(),
                            ).results()

                            objs = []

                            # Prepare objects for bulk create
                            for item in contract_items:

                                cont = Contract.objects.get(
                                    contract_id=contract["contract_id"]
                                )
                                itm = EveType.objects.get(pk=item["type_id"])

                                contract_item = ContractItem(
                                    contract=cont,
                                    eve_type=itm,
                                    quantity=item["quantity"],
                                )

                                objs.append(contract_item)

                            try:
                                ContractItem.objects.bulk_create(objs)
                                logger.debug(
                                    "Succesfully added items for contract %s into database"
                                    % contract["contract_id"]
                                )
                            except Error as e:
                                logger.error(
                                    "Error adding items for contract %s: %s"
                                    % (contract["contract_id"], e)
                                )

                            """
                            Notifications for users who have the notifications enabled
                            """

                            disablation_hint = "*You can disable these notifications from your program settings*"

                            user_message = {
                                "title": "New contract assigned",
                                "description": "A new buyback contract {0} with a value of {1} ISK has been assigned to you.\n\n{2}".format(
                                    tracking.tracking_number,
                                    intcomma(int(contract["price"])),
                                    disablation_hint,
                                ),
                                "footer": "AA Buyback Program",
                            }

                            # If tracking is active and we should send a message for our users
                            if tracking.program.discord_dm_notification:

                                send_user_notification(
                                    user=self.user,
                                    level="success",
                                    message=user_message,
                                )

                            # Notifications for the discord channel
                            if tracking.program.discord_channel_notification:
                                send_message_to_discord_channel(
                                    channel_id=tracking.program.discord_channel_notification,
                                    message=user_message,
                                )

                        # If contract was updated instead of created
                        else:
                            logger.debug("Contract %s updated." % obj.contract_id)

                        # Check if the contract status has changed from ongoing to finished.
                        if (
                            old_contract.status == "outstanding"
                            and obj.status == "finished"
                        ):
                            # Get notification settings for the contract issuer
                            user_settings = UserSettings.objects.get(
                                user=tracking.issuer_user
                            )

                            # If user has not disabled notifications
                            if user_settings.disable_notifications is False:

                                disablation_hint = "*You can disable these notifications from [My Settings]({0}/buybackprogram/user_settings_edit)*".format(
                                    get_site_url()
                                )

                                # Check if the contract was accepted or rejected
                                if contract["status"] == "finished":
                                    status = "accepted"
                                elif contract["status"] == "rejected":
                                    status = "rejected"
                                else:
                                    status = contract["status"]

                                user_message = {
                                    "title": "Buyback contract {0}".format(status),
                                    "description": "Your outstanding buyback contract {0} with a value of {1} ISK has been {2} by {3}.\n\n{4}".format(
                                        tracking.tracking_number,
                                        intcomma(int(contract["price"])),
                                        status,
                                        self.character.character,
                                        disablation_hint,
                                    ),
                                    "footer": "AA Buyback Program",
                                }

                                send_user_notification(
                                    user=tracking.issuer_user,
                                    level="success",
                                    message=user_message,
                                )

                    break  # If we have found a match from our ESI contracts wi will stop looping on the contracts

    @fetch_token_for_owner(["esi-contracts.read_character_contracts.v1"])
    def _fetch_contracts(self, token) -> list:

        character_id = self.character.character.character_id

        contracts = esi.client.Contracts.get_characters_character_id_contracts(
            character_id=character_id,
            token=token.valid_access_token(),
        ).results()

        return contracts

    @fetch_token_for_owner(["esi-contracts.read_corporation_contracts.v1"])
    def _fetch_corporation_contracts(self, token) -> list:

        corporation_id = self.character.character.corporation_id

        contracts = esi.client.Contracts.get_corporations_corporation_id_contracts(
            corporation_id=corporation_id,
            token=token.valid_access_token(),
        ).results()

        return contracts

    def token(self, scopes=None) -> Tuple[Token, int]:
        """returns a valid Token for the owner"""
        token = None
        error = None

        # abort if character is not configured
        if self.character is None:
            logger.error("%s: No character configured to sync", self)
            error = self.ERROR_NO_CHARACTER

        # abort if character does not have sufficient permissions
        elif self.corporation and not self.character.user.has_perm(
            "buybackprogram.manage_programs"
        ):
            logger.error(
                "%s: This character does not have sufficient permission to sync contracts",
                self,
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        # abort if character does not have sufficient permissions
        elif not self.character.user.has_perm("buybackprogram.manage_programs"):
            logger.error(
                "%s: This character does not have sufficient permission to sync contracts",
                self,
            )
            error = self.ERROR_INSUFFICIENT_PERMISSIONS

        else:
            try:
                # get token
                token = (
                    Token.objects.filter(
                        user=self.character.user,
                        character_id=self.character.character.character_id,
                    )
                    .require_scopes(scopes)
                    .require_valid()
                    .first()
                )
            except TokenInvalidError:
                logger.error("%s: Invalid token for fetching calendars", self)
                error = self.ERROR_TOKEN_INVALID
            except TokenExpiredError:
                logger.error("%s: Token expired for fetching calendars", self)
                error = self.ERROR_TOKEN_EXPIRED
            else:
                if not token:
                    logger.error("%s: No token found with sufficient scopes", self)
                    error = self.ERROR_TOKEN_INVALID

        return token, error

    def __str__(self):
        return (
            self.character.character.character_name
            + " ["
            + self.corporation.corporation_ticker
            + "]"
        )


class Location(models.Model):
    """ Location where the buyback program is operated at """

    name = models.CharField(
        max_length=32, help_text="Structure name where the contracts are accepted at"
    )

    eve_solar_system = models.ForeignKey(
        EveSolarSystem,
        verbose_name="Solar system",
        help_text="System where the buyback structure is located",
        blank=True,
        default=None,
        null=True,
        on_delete=models.deletion.SET_DEFAULT,
        related_name="+",
    )

    owner = models.ForeignKey(
        Owner,
        verbose_name="Manager",
        help_text="Player managing this location",
        null=True,
        on_delete=models.deletion.CASCADE,
    )

    structure_id = models.BigIntegerField(
        verbose_name="Ingame unique ID for structure",
        default=None,
        blank=True,
        null=True,
        help_text="The ID for the structure you wish to accept the contracts at. If left empty the program statistics page will not track if the contract is actually made at the correct structure or not. To get the ID for the structure see readme for getting structure IDs",
    )

    def __str__(self):

        return (
            self.eve_solar_system.name
            + " | "
            + self.name
            + ", ID: "
            + str(self.structure_id)
        )


class Program(models.Model):
    """An Eve Online buyback program"""

    owner = models.ForeignKey(
        Owner,
        verbose_name="Manager",
        help_text="Character that is used to manage this program.",
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    is_corporation = models.BooleanField(
        default=False,
        help_text="If we should use the corporation of the manager as the contract receiver instead of the character.",
    )

    location = models.ForeignKey(
        Location,
        help_text="The location where contracts should be created at.",
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )

    tax = models.IntegerField(
        verbose_name="Default tax",
        default=0,
        blank=False,
        null=False,
        help_text="A default tax rate in this program that is applied on all items.",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    hauling_fuel_cost = models.IntegerField(
        verbose_name="Hauling fuel cost per m³",
        default=0,
        help_text="ISK per m³ that will be removed from the buy price ie. to cover jump freighet fuel costs. <b>Should not be used with price dencity modifier</b>",
    )

    price_dencity_modifier = models.BooleanField(
        verbose_name="Price density modifier",
        default=False,
        help_text="Should we modify buy prices for items with high volume and low value ie. T1 industrial hulls. <b>Should not be used with hauling fuel cost</b>",
    )

    price_dencity_treshold = models.IntegerField(
        verbose_name="Price density threshold",
        default=0,
        null=True,
        help_text="At what ISK/m3 do we start to apply the low isk dencity tax. Tritanium is 500 ISK/m³ @ 5 ISK per unit price. PLEX is 14,5Trillion ISK/m³ @2.9M per unit price.",
    )

    price_dencity_tax = models.IntegerField(
        verbose_name="Price density tax",
        default=0,
        null=True,
        help_text="How much tax do we apply on the low isk density items.",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    allow_all_items = models.BooleanField(
        default=True,
        help_text="If true all items are accepted to the buyback program. You can set extra taxes or disallow individual items from the program item section. If set to false you need to add each accepted item into the program item section. Blueprints are not included in all items.",
    )

    use_refined_value = models.BooleanField(
        verbose_name="Ore: Use refined value",
        default=False,
        help_text="Take refined value into account when calculating prices for ore, ice and moon goo",
    )

    use_compressed_value = models.BooleanField(
        verbose_name="Ore: Use compressed value",
        default=False,
        help_text="Take compressed value into account when calculating prices for ore, ice and moon goo",
    )

    use_raw_ore_value = models.BooleanField(
        verbose_name="Ore: Use raw value",
        default=True,
        help_text="Take raw ore value into account when calculating prices for ore, ice and moon goo",
    )

    allow_unpacked_items = models.BooleanField(
        verbose_name="Allow unpacked items",
        default=False,
        help_text="Do you want to allow unpacked items in this program such as assembled ship hulls?",
    )

    refining_rate = models.IntegerField(
        verbose_name="Refining rate",
        default=0,
        null=True,
        help_text="Refining rate to be used if ore refined value is active",
        validators=[MaxValueValidator(100), MinValueValidator(0)],
    )

    blue_loot_npc_price = models.BooleanField(
        verbose_name="NPC price for: Sleeper Components",
        default=False,
        help_text="Use NPC price as value for blue loot",
    )

    red_loot_npc_price = models.BooleanField(
        verbose_name="NPC price for: Triglavian Survey Database",
        default=False,
        help_text="Use NPC price as value for red loot",
    )

    ope_npc_price = models.BooleanField(
        verbose_name="NPC price for: Overseer's Personal Effects",
        default=False,
        help_text="Use NPC price as value for OPEs",
    )

    restricted_to_group = models.ManyToManyField(
        Group,
        blank=True,
        related_name="buybackprogram_require_groups",
        help_text="The group(s) that will be able to see this buyback program. If none is selected program is open for all.",
    )
    restricted_to_state = models.ManyToManyField(
        State,
        blank=True,
        related_name="buybackprogram_require_states",
        help_text="The state(s) that will be able to see this buyback program. If none is selected program is open for all.",
    )

    discord_dm_notification = models.BooleanField(
        verbose_name="Discord direct messages for new contracts",
        default=False,
        help_text="Check if you want to receive a direct message notification each time a new contract is created. <b>Requires aa-discordbot app or discordproxy app to work</b>",
    )

    discord_channel_notification = models.BigIntegerField(
        verbose_name="Discord channel ID for notifications",
        null=True,
        blank=True,
        help_text="Check if you want to feed new contracts to a discord channel. <b>Requires aa-discordbot app or discordproxy app to work</b>",
    )

    class Meta:
        default_permissions = ()

    def clean(self):
        super().clean()
        if (
            self.allow_all_items
            and not self.use_refined_value
            and not self.use_compressed_value
            and not self.use_raw_ore_value
        ):
            raise ValidationError(
                "All items are allowed but not a single pricing method for ores is selected. Please use at least one pricing method for ores if all items is allowed."
            )
        if self.price_dencity_modifier and not self.price_dencity_tax:
            raise ValidationError(
                "Price density is used but value for price density tax is missing"
            )
        if self.price_dencity_modifier and not self.price_dencity_treshold:
            raise ValidationError(
                "Price density is used but value for price density threshold is missing"
            )
        if self.use_refined_value and not self.refining_rate:
            raise ValidationError(
                "Refined value is used for ore pricing method but no refining rate is provided. Provide a refining rate to used with this pricing model."
            )


class ProgramItem(models.Model):
    """Items in the buyback program for a corp"""

    program = models.ForeignKey(
        Program,
        on_delete=models.deletion.CASCADE,
        help_text="What program do these items belong to",
    )
    item_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Select item for special tax",
    )
    item_tax = models.IntegerField(
        verbose_name="Item tax adjustment",
        default=0,
        null=True,
        help_text="How much do you want to adjust the default tax on this item. Can be a positive or a negative value.",
        validators=[MaxValueValidator(100), MinValueValidator(-100)],
    )

    disallow_item = models.BooleanField(
        verbose_name="Disallow item in program",
        default=False,
        help_text="You can disallow an item from a buyback location. It will return 0 price if disallowed.",
    )

    class Meta:
        default_permissions = ()
        unique_together = ["program", "item_type"]


class ItemPrices(models.Model):
    eve_type = models.OneToOneField(
        EveType,
        on_delete=models.deletion.CASCADE,
        unique=True,
    )
    buy = models.BigIntegerField()
    sell = models.BigIntegerField()
    updated = models.DateTimeField()


class Tracking(models.Model):
    program = models.ForeignKey(
        Program,
        null=True,
        on_delete=models.deletion.SET_NULL,
        related_name="+",
    )
    issuer_user = models.ForeignKey(
        User,
        on_delete=models.deletion.CASCADE,
        related_name="+",
    )
    value = models.BigIntegerField(null=False)
    taxes = models.BigIntegerField(null=False)
    hauling_cost = models.BigIntegerField(null=False)
    donation = models.BigIntegerField(null=True, blank=True)
    net_price = models.BigIntegerField(null=False)
    tracking_number = models.CharField(max_length=20)


class TrackingItem(models.Model):

    tracking = models.ForeignKey(
        Tracking,
        on_delete=models.deletion.CASCADE,
        help_text="What tracking do these items belong to",
    )

    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Item type information",
    )

    buy_value = models.BigIntegerField(null=False)

    quantity = models.IntegerField()


class Contract(models.Model):

    assignee_id = models.IntegerField()
    availability = models.CharField(max_length=20)
    contract_id = models.IntegerField()
    date_completed = models.DateTimeField(null=True)
    date_expired = models.DateTimeField(null=True)
    date_issued = models.DateTimeField()
    for_corporation = models.BooleanField()
    issuer_corporation_id = models.IntegerField()
    issuer_id = models.IntegerField()
    start_location_id = models.BigIntegerField(null=True)
    price = models.BigIntegerField()
    status = models.CharField(max_length=30)
    title = models.CharField(max_length=128)
    volume = models.BigIntegerField()


class ContractItem(models.Model):

    contract = models.ForeignKey(
        Contract,
        on_delete=models.deletion.CASCADE,
        help_text="What contract do these items belong to",
    )

    eve_type = models.ForeignKey(
        EveType,
        on_delete=models.deletion.CASCADE,
        help_text="Item type information",
    )

    quantity = models.IntegerField()


class UserSettings(models.Model):
    """
    User settings
    """

    user = models.ForeignKey(
        User,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET(get_sentinel_user),
    )

    disable_notifications = models.BooleanField(
        default=False,
    )

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("User Settings")
        verbose_name_plural = _("User Settings")



from django.conf import settings

from .utils import clean_setting

# put your app settings here


EXAMPLE_SETTING_ONE = getattr(settings, "EXAMPLE_SETTING_ONE", None)


# Hard timeout for tasks in seconds to reduce task accumulation during outages
BUYBACKPROGRAM_TASKS_TIME_LIMIT = clean_setting("BUYBACKPROGRAM_TASKS_TIME_LIMIT", 7200)

# Tracking number tag
BUYBACKPROGRAM_TRACKING_PREFILL = clean_setting(
    "BUYBACKPROGRAM_TRACKING_PREFILL", "aa-bbp"
)


def allianceauth_discordbot_active():
    """
    check if allianceauth-dicordbot is installed and active
    :return:
    """
    return "aadiscordbot" in settings.INSTALLED_APPS



from allianceauth.services.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def note_missing_jita_buy(buy_price, name):
    if buy_price == 0:
        note = {
            "icon": "fa-question",
            "color": "red",
            "message": "%s has no buy orders in jita" % name,
        }
        return note
    else:
        return False


def note_price_dencity_tax(name, price_dencity, price_dencity_tax):
    if price_dencity_tax:
        note = {
            "icon": "fa-weight-hanging",
            "color": "orange",
            "message": "%s has price dencity of %s isk/m³, %s %s low price dencity tax applied"
            % (name, price_dencity, price_dencity_tax, "%"),
        }

        return note
    else:
        return False


def note_item_disallowed(item_disallowed, name):
    if item_disallowed:
        note = {
            "icon": "fa-hand-paper",
            "color": "red",
            "message": "%s is disallowed in this program" % name,
        }
        return note
    else:
        return False


def note_unpublished_item(name):
    note = {
        "icon": "fa-hand-paper",
        "color": "red",
        "message": "%s is not a published item. Special commondite or expired item?"
        % name,
    }
    return note


def note_missing_typematerials(type_materials, name):
    if type_materials.count() == 0:

        logger.error(
            "TypeMaterials not found for %s. Did you forget to run buybackprogram_load_data?"
            % name
        )

        note = {
            "icon": "fa-exclamation-triangle",
            "color": "red",
            "message": "Refined price valuation is used in this program but TypeMaterials for %s are missing from the database."
            % name,
        }
        return note
    else:
        return False


def note_item_specific_tax(name, item_tax):
    if item_tax:
        if item_tax > 0:
            note = {
                "icon": "fa-percentage",
                "color": "orange",
                "message": "%s has an additional %s %s item spesific tax applied on it"
                % (name, item_tax, "%"),
            }
            return note
        else:
            note = {
                "icon": "fa-percentage",
                "color": "green",
                "message": "%s has an decreased %s %s item spesific tax applied on it"
                % (name, item_tax, "%"),
            }
        return note
    else:
        return False


def note_no_price_data(name):
    note = {
        "icon": "fa-question",
        "color": "red",
        "message": "%s has no price data" % name,
    }
    return note


def note_refined_price_used(name):
    note = {
        "icon": "fa-industry",
        "color": "#5858df",
        "message": "Best price: Using refined price for %s" % name,
    }

    return note


def note_compressed_price_used(name):
    note = {
        "icon": "fa-file-archive",
        "color": "#5858df",
        "message": "Best price: Using compressed price for %s" % name,
    }
    return note


def note_npc_price(name):
    note = {
        "icon": "fa-robot",
        "color": "#5858df",
        "message": "Using NPC buy price for %s instead of Jita buy prices" % name,
    }
    return note



from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.utils.html import format_html
from eveuniverse.models import EveMarketGroup, EveType

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.forms import ProgramItemForm, ProgramMarketGroupForm
from buybackprogram.models import Program, ProgramItem
from buybackprogram.utils import messages_plus

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def program_special_taxes(request, program_pk):
    program = Program.objects.get(pk=program_pk)
    program_items = ProgramItem.objects.filter(program=program)

    context = {
        "program": program,
        "program_items": program_items,
    }

    return render(request, "buybackprogram/program_special_taxes.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def program_edit_item(request, program_pk):
    program = Program.objects.get(pk=program_pk)

    program_items = ProgramItem.objects.filter(program=program)

    if request.method != "POST":
        form = ProgramItemForm()
    else:
        form = ProgramItemForm(
            request.POST,
            value=int(request.POST["item_type"]),
        )

        if form.is_valid():

            item_tax = form.cleaned_data["item_tax"]
            disallow_item = form.cleaned_data["disallow_item"]

            item_type = form.cleaned_data["item_type"]

            ProgramItem.objects.update_or_create(
                item_type=item_type,
                program=program,
                defaults={
                    "item_tax": item_tax,
                    "disallow_item": disallow_item,
                },
            )

            if disallow_item:
                messages_plus.warning(
                    request,
                    format_html(
                        "Disallowed <strong>{}</strong> in program",
                        item_type.name,
                    ),
                )

            else:
                messages_plus.success(
                    request,
                    format_html(
                        "Adjusted <strong>{}</strong> tax in program with <strong>{}</strong> %, tax is now set at {} %",
                        item_type.name,
                        item_tax,
                        program.tax + item_tax,
                    ),
                )

            return HttpResponseRedirect(request.path_info)

    context = {
        "program": program,
        "program_items": program_items,
        "form": form,
    }

    return render(request, "buybackprogram/program_edit_item.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def program_edit_marketgroup(request, program_pk):
    program = Program.objects.get(pk=program_pk)

    program_items = ProgramItem.objects.filter(program=program)

    if request.method != "POST":
        form = ProgramMarketGroupForm()
    else:
        form = ProgramMarketGroupForm(
            request.POST,
            value=int(request.POST["marketgroup"]),
        )

        if form.is_valid():

            item_types = []
            item_count = 0

            item_tax = form.cleaned_data["item_tax"]
            disallow_item = form.cleaned_data["disallow_item"]

            marketgroups = EveMarketGroup.objects.get(
                pk=form.cleaned_data["marketgroup"].id
            )

            item_type = EveType.objects.filter(
                eve_market_group=form.cleaned_data["marketgroup"]
            )

            item_types.append(item_type)

            for m in marketgroups.market_group_children.all():

                item_type = EveType.objects.filter(eve_market_group=m)

                item_types.append(item_type)

                for m2 in m.market_group_children.all():

                    item_type = EveType.objects.filter(eve_market_group=m2)

                    item_types.append(item_type)

            for sub_types in item_types:
                for item in sub_types:

                    logger.debug("Adjusting tax for %s" % item)

                    ProgramItem.objects.update_or_create(
                        item_type=item,
                        program=program,
                        defaults={
                            "item_tax": item_tax,
                            "disallow_item": disallow_item,
                        },
                    )

                    item_count += 1

            if disallow_item:
                messages_plus.warning(
                    request,
                    format_html(
                        "Disallowed <strong>{}</strong> items from market group {} in program",
                        item_count,
                        form.cleaned_data["marketgroup"],
                    ),
                )
            else:
                messages_plus.success(
                    request,
                    format_html(
                        "Added <strong>{}</strong> items from market group {} to program with <strong>{}</strong> % tax",
                        item_count,
                        form.cleaned_data["marketgroup"],
                        item_tax,
                    ),
                )

            return HttpResponseRedirect(request.path_info)

    context = {
        "program": program,
        "program_items": program_items,
        "form": form,
    }

    return render(request, "buybackprogram/program_edit_marketgroup.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_item_remove(request, item_pk, program_pk):

    program_item = ProgramItem.objects.get(item_type=item_pk)

    name = program_item.item_type

    program_item.delete()

    messages_plus.warning(
        request,
        format_html(
            "Deleted <strong>{}</strong> from program",
            name,
        ),
    )

    return redirect("buybackprogram:program_special_taxes", program_pk)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_item_remove_all(request, program_pk):

    program_item = ProgramItem.objects.all()

    program_item.delete()

    messages_plus.warning(
        request,
        format_html(
            "Deleted all special taxation items from program",
        ),
    )

    return redirect("buybackprogram:program_special_taxes", program_pk)



from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.shortcuts import redirect, render
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from ..models import Contract, ContractItem, Tracking, TrackingItem

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def my_stats(request):

    valid_contracts = []

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    tracking_numbers = Tracking.objects.all()

    for tracking in tracking_numbers:

        contract = Contract.objects.filter(
            issuer_id__in=characters, title__contains=tracking.tracking_number
        ).first()

        if contract:

            contract.tracking = tracking

            logger.debug(
                "Contract %s has a match in tracking numbers" % contract.contract_id
            )

            if contract.status == "outstanding":
                values["outstanding"] += contract.price
                values["outstanding_count"] += 1
            if contract.status == "finished":
                values["finished"] += contract.price
                values["finished_count"] += 1

            contract.notes = []

            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            logger.debug("Issuer name for contract is %s" % contract.issuer_name)

            contract.items = ContractItem.objects.filter(contract=contract)

            if tracking.program:
                structure_id = tracking.program.location.structure_id
            else:
                structure_id = False

            if structure_id and structure_id != contract.start_location_id:
                note = {
                    "icon": "fa-compass",
                    "color": "red",
                    "message": "Contract location id %s does not match program location id %s"
                    % (contract.start_location_id, structure_id),
                }

                contract.notes.append(note)

            if tracking.net_price != contract.price:
                note = {
                    "icon": "fa-skull-crossbones",
                    "color": "red",
                    "message": "Tracked price does not match contract price. Either a mistake in the copy pasted values or seller has changed contracted items after calculating the price",
                }

                contract.note = note

            if not tracking.tracking_number == contract.title:
                note = {
                    "icon": "fa-exclamation",
                    "color": "orange",
                    "message": "Contract description contains extra characterse besides the tracking number. The description should be: '%s', instead it is: '%s'"
                    % (tracking.tracking_number, contract.title),
                }

                contract.notes.append(note)

            valid_contracts.append(contract)

    context = {
        "contracts": valid_contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_stats(request):

    valid_contracts = []

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    logger.debug("Got characters for manager: %s" % characters)

    corporations = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__corporation_id", flat=True
    )

    logger.debug("Got corporations for manager: %s" % corporations)

    tracking_numbers = Tracking.objects.all()

    for tracking in tracking_numbers:

        contract = Contract.objects.filter(
            Q(assignee_id__in=characters) | Q(assignee_id__in=corporations),
            title__contains=tracking.tracking_number,
        ).first()

        if contract:

            contract.tracking = tracking

            logger.debug(
                "Contract %s has a match in tracking numbers" % contract.contract_id
            )

            if contract.status == "outstanding":
                values["outstanding"] += contract.price
                values["outstanding_count"] += 1
            if contract.status == "finished":
                values["finished"] += contract.price
                values["finished_count"] += 1

            contract.notes = []

            try:
                issuer_character = CharacterOwnership.objects.get(
                    character__character_id=contract.issuer_id
                )
                logger.debug(
                    "Got issuer character from auth: %s" % issuer_character.user
                )

            except CharacterOwnership.DoesNotExist:
                issuer_character = False
                logger.debug("Contract issuer not registered on AUTH")

                note = {
                    "icon": "fa-question",
                    "color": "orange",
                    "message": "Issuer not registered on AUTH. Possibly an unregistered alt.",
                }

                contract.notes.append(note)

            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

            contract.items = ContractItem.objects.filter(contract=contract)

            if tracking.program:
                structure_id = tracking.program.location.structure_id
            else:
                structure_id = False

            if structure_id and structure_id != contract.start_location_id:
                note = {
                    "icon": "fa-compass",
                    "color": "red",
                    "message": "Contract location id %s does not match program location id %s"
                    % (contract.start_location_id, structure_id),
                }

                contract.notes.append(note)

            if tracking.net_price != contract.price:
                note = {
                    "icon": "fa-skull-crossbones",
                    "color": "red",
                    "message": "Tracked price for %s does not match contract price. See details for more information"
                    % tracking.tracking_number,
                }

                contract.notes.append(note)

            if (
                contract.assignee_id in corporations
                and not tracking.program.is_corporation
            ):
                note = {
                    "icon": "fa-home",
                    "color": "orange",
                    "message": "Contract %s is made for your corporation while they should be made directly to your character in this program."
                    % tracking.tracking_number,
                }

                contract.notes.append(note)

            if tracking.program:
                if (
                    contract.assignee_id not in corporations
                    and tracking.program.is_corporation
                ):
                    note = {
                        "icon": "fa-user",
                        "color": "orange",
                        "message": "Contract %s is made for your character while they should be made to your corporation in this program."
                        % tracking.tracking_number,
                    }

                    contract.notes.append(note)

            if not tracking.tracking_number == contract.title:
                note = {
                    "icon": "fa-exclamation",
                    "color": "orange",
                    "message": "Contract description contains extra characterse besides the tracking number. The description should be: '%s', instead it is: '%s'"
                    % (tracking.tracking_number, contract.title),
                }

                contract.notes.append(note)

            valid_contracts.append(contract)

    context = {
        "contracts": valid_contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/program_stats.html", context)


@login_required
@permission_required("buybackprogram.see_all_statics")
def program_stats_all(request):

    valid_contracts = []

    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    tracking_numbers = Tracking.objects.all()

    for tracking in tracking_numbers:

        contract = Contract.objects.filter(
            title__contains=tracking.tracking_number
        ).first()

        if contract:

            contract.tracking = tracking

            logger.debug(
                "Contract %s has a match in tracking numbers" % contract.contract_id
            )

            if contract.status == "outstanding":
                values["outstanding"] += contract.price
                values["outstanding_count"] += 1
            if contract.status == "finished":
                values["finished"] += contract.price
                values["finished_count"] += 1

            contract.notes = []

            try:
                issuer_character = CharacterOwnership.objects.get(
                    character__character_id=contract.issuer_id
                )
                logger.debug(
                    "Got issuer character from auth: %s" % issuer_character.user
                )

            except CharacterOwnership.DoesNotExist:
                issuer_character = False
                logger.debug("Contract issuer not registered on AUTH")

                note = {
                    "icon": "fa-question",
                    "color": "orange",
                    "message": "Issuer not registered on AUTH. Possibly an unregistered alt.",
                }

                contract.notes.append(note)

            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

            contract.items = ContractItem.objects.filter(contract=contract)

            if tracking.program:
                structure_id = tracking.program.location.structure_id
            else:
                structure_id = False

            if structure_id and structure_id != contract.start_location_id:
                note = {
                    "icon": "fa-compass",
                    "color": "red",
                    "message": "Contract location id %s does not match program location id %s"
                    % (contract.start_location_id, structure_id),
                }

                contract.notes.append(note)

            if tracking.net_price != contract.price:
                note = {
                    "icon": "fa-skull-crossbones",
                    "color": "red",
                    "message": "Tracked price for %s does not match contract price. See details for more information"
                    % tracking.tracking_number,
                }

                contract.notes.append(note)

            if not tracking.tracking_number == contract.title:
                note = {
                    "icon": "fa-exclamation",
                    "color": "orange",
                    "message": "Contract description contains extra characterse besides the tracking number. The description should be: '%s', instead it is: '%s'"
                    % (tracking.tracking_number, contract.title),
                }

                contract.notes.append(note)

            valid_contracts.append(contract)

    context = {
        "contracts": valid_contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/program_stats_all.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def contract_details(request, contract_title):
    try:

        notes = []

        contract = Contract.objects.get(title__contains=contract_title)

        contract_items = ContractItem.objects.filter(contract=contract)

        tracking = Tracking.objects.get(
            tracking_number=contract_title,
        )

        tracking_items = TrackingItem.objects.filter(tracking=tracking)

        contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)
        contract.assignee_name = EveEntity.objects.resolve_name(contract.assignee_id)

        if contract.price != tracking.net_price:
            note = {
                "icon": "fa-skull-crossbones",
                "color": "alert-danger",
                "message": "Tracked price does not match contract price. You have either made an mistake in the tracking number or the contract price copy paste. Please remake contract.",
            }
            notes.append(note)

        context = {
            "notes": notes,
            "contract": contract,
            "contract_items": contract_items,
            "tracking": tracking,
            "tracking_items": tracking_items,
        }

        return render(request, "buybackprogram/contract_details.html", context)

    except Contract.DoesNotExist:
        return redirect("buybackprogram/stats.html")
